// OpenCV related
#include <opencv2/opencv.hpp>
#include <opencv2/core.hpp>
#include <opencv2/highgui.hpp>

// ArUco marker related
#include <opencv2/aruco.hpp>

#include <math.h>
#include <Eigen/Core>
#include <Eigen/Geometry>

// ROS related
#include <ros/ros.h>
#include <ros/node_handle.h>
#include <cv_bridge/cv_bridge.h>
#include <image_transport/image_transport.h>
#include <sensor_msgs/Image.h>
#include <sensor_msgs/CompressedImage.h>

#define PI 3.14159265359

class arMarkerRecognizer{
public:
	ros::NodeHandle nh_;
	image_transport::ImageTransport it_;
	image_transport::Publisher image_pub_;

	static const int cnt = 0;
	// Horizontal image sensor size.
	float h_dis = 2.9;
	// Vertical image sensor size.
	float v_dis = 2.2;
	// Horizontal image size.
	int h_size = 0;
	// Vertical image size.
	int v_size = 0;
	// Field of view.
	float fov = 68;
	// Distance to center of ArUco marker in image.(X, Y)
	float image_dis_x = 0.0;
	float image_dis_y = 0.0;
	// Distance to center in ArUco marker.(X, Y)
	float aruco_dis_x = 0.0;
	float aruco_dis_y = 0.0;
	bool mode = false;

	arMarkerRecognizer()
		: it_(nh_)
	{
		image_pub_ = it_.advertise("/camera/image_raw", 10);
	}

	void recognizeArMarker(cv::Mat frame)
	{
		cv::Ptr<cv::aruco::DetectorParameters> parameters = cv::aruco::DetectorParameters::create();
		cv::Ptr<cv::aruco::Dictionary> dictionary = cv::aruco::getPredefinedDictionary(cv::aruco::DICT_4X4_100);
		// Coordinates of center of the ArUco marker.
		int corner_id = 0;
		std::string text = "X: -nan, Y: -nan, Z: -nan";

		for(int i = 0; i < cnt + 1; i++){
			std::vector<std::vector<cv::Point2f>> corners, rejectedCandidates;
			std::vector<int> ids;

			// Detection of ArUco markers.
			cv::aruco::detectMarkers(frame, dictionary, corners, ids, parameters, rejectedCandidates);
	
			if(ids.size() > 0){
				// Draw detection result of ArUco marker.
				cv::aruco::drawDetectedMarkers(frame, corners, ids);

				aruco_dis_x = maxCornerCoordinate(corners[i][0].x, corners[i][1].x, corners[i][2].x, corners[i][3].x, corner_id);
				if(corner_id == 0){
					aruco_dis_x = std::max(aruco_dis_x - corners[i][3].x, aruco_dis_x - corners[i][corner_id + 1].x);
				} else if(corner_id == 3){
					aruco_dis_x = std::max(aruco_dis_x - corners[i][0].x, aruco_dis_x - corners[i][corner_id - 1].x);
				} else {
					aruco_dis_x = std::max(aruco_dis_x - corners[i][corner_id - 1].x, aruco_dis_x - corners[i][corner_id + 1].x);
				}

				aruco_dis_y = maxCornerCoordinate(corners[i][0].y, corners[i][1].y, corners[i][2].y, corners[i][3].y, corner_id);
				if(corner_id == 0){
					aruco_dis_y = std::max(aruco_dis_y - corners[i][3].y, aruco_dis_y - corners[i][corner_id + 1].y);
				} else if(corner_id == 3){
					aruco_dis_y = std::max(aruco_dis_y - corners[i][0].y, aruco_dis_y - corners[i][corner_id - 1].y);
				} else {
					aruco_dis_y = std::max(aruco_dis_y - corners[i][corner_id - 1].y, aruco_dis_y - corners[i][corner_id + 1].y);
				}

				image_dis_x = std::min({corners[0][0].x, corners[0][1].x, corners[0][2].x, corners[0][3].x}) + aruco_dis_x / 2.0;
				image_dis_y = std::min({corners[0][0].y, corners[0][1].y, corners[0][2].y, corners[0][3].y}) + aruco_dis_y / 2.0;

				Eigen::Vector3f cnvCameraToWorld = convertCameraToWorldCoordinates(mode);
				text = checkNan(cnvCameraToWorld);
			}
			cv::putText(frame, text, cv::Point(50, 700), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 255, 0), 3, cv::LINE_4);
		}
	}

	/* Description: Diagonal distance of image sensor (mm).
	 * Argument: float h_dis: Image sensor horizontal distance (mm).
	 *           float v_dis: Image sensor vertical distance (mm).
	 *   Return: float diagonal_distance: Diagonal distance of image sensor (mm).
	 */
	float imageSensorDiagonalDistance(float h_dis, float v_dis)
	{
		float diagonal_distance;
		diagonal_distance = std::sqrt(std::pow(h_dis, 2) + std::pow(v_dis, 2));

		return diagonal_distance;
	}

	/* Description: Get focal length(mm).
	 * Argument: float dis: Image sensor length.
	 *           float fov: Field of view(degree).
	 *   Return: float pixel_: Focal length (mm).
	 */
	float focalLength(float dis, float fov, const bool debug=false)
	{
		float f_;
		// Convert fov from degrees to radians.
		fov = fov / 180.0 * PI;
		f_ = dis / (2.0 * std::tan(fov / 2.0));

		return f_;
	}

	/* Description: Convert focal length from mm to pixel.
	 * Argument: float dis: Image sensor length (mm).
	 *           float size: Image size (pixel).
	 *   Return: float pixel_: Focal length (pixel).
	 */
	float focalLengthMmToPixels(float dis, int size, const bool debug=false)
	{
		float pixel_;
		float diagonal_dis;
		diagonal_dis = imageSensorDiagonalDistance(h_dis, v_dis);
		pixel_ = size / dis * focalLength(diagonal_dis, fov, mode);

		return pixel_;
	}

	/* Description: Camera internal parameters.
	 */
	Eigen::Matrix3f cameraInternalParam(const bool debug=false)
	{
		Eigen::Matrix3f m;

		m << focalLengthMmToPixels(h_dis, h_size, mode), 0, h_size / 2,
				0, focalLengthMmToPixels(v_dis, v_size, mode), v_size / 2,
				0, 0, 1;
		
		return m;
	}

	/* Description: Convert from image coordinate to camera coordinate.
	 * Return: Eigen::Vector3f v_: Camera coordinates.
	 */
	Eigen::Vector3f convertImageToCameraCoordinates(const bool debug=false)
	{
		Eigen::Vector3f v_;
		v_.x() = (image_dis_x - h_size / 2) / focalLengthMmToPixels(h_dis, h_size, mode);
		v_.y() = (image_dis_y - v_size / 2) / focalLengthMmToPixels(v_dis, v_size, mode);
		v_.z() = 1.0;

		return v_;
	}

	/* Description: Convert from camera coordinate to world coordinate.
	 * Return: Eigen::Vector3f v_: World coordinates.
	 */
	Eigen::Vector3f convertCameraToWorldCoordinates(const bool debug=false)
	{
		Eigen::Vector3f v_, v_camera, t, tmp;
		Eigen::Matrix3f R, Rinv;
		t << 0, 0, 0;
		R << 1, 0, 0, 0, 1, 0, 0, 0, 1;
		Rinv = R.inverse();
		tmp = convertImageToCameraCoordinates(mode);
		float marker_size = 0.04;
		// Focal length d'.
		float dd = focalLengthMmToPixels(v_dis, v_size, mode);
		// The size of the subject reflected on the image sensor.
		float hd = aruco_dis_y;
		// Distance to the subject.
		float d;
		// Size of the subject is 0.04m(4cm).
		float h = 0.04;

		v_camera.x() = -tmp.x() * (h_size / 2.0 - image_dis_x) * (marker_size / aruco_dis_x) / tmp.x();
		v_camera.y() = (tmp.z() * h * dd) / hd;
		v_camera.z() = tmp.y() * (v_size / 2.0 - image_dis_y) * (marker_size / aruco_dis_y) / tmp.y();

		v_ = Rinv * (v_camera - t);

		return v_;
	}

	/* Description: Maximum value of four corners in ArUco marker.
	 * Return: float max_val: Maximum value.
	 */
	int maxCornerCoordinate(float corner0, float corner1, float corner2, float corner3, int& corner_id)
	{
		float max_val = 0.0;

		if(max_val < corner0){
			max_val = corner0;
			corner_id = 0;
		}
		if(max_val < corner1){
			max_val = corner1;
			corner_id = 1;
		}
		if(max_val < corner2){
			max_val = corner2;
			corner_id = 2;
		}
		if(max_val < corner3){
			max_val = corner3;
			corner_id = 3;
		}

		return max_val;
	}

	/* Description: Check for NAN.
	 * Return: std::string text: The text to display on the image.
	 */
	std::string checkNan(Eigen::Vector3f& num)
	{
		std::string text;
		if(std::isnan(num.x())){
			text = "X: -nan, ";
		} else {
			text = "X: " + std::to_string(num.x()) + ", ";
		}
		if(std::isnan(num.y())){
			text += "Y: -nan, ";
		} else {
			text += "Y: " + std::to_string(num.y()) + ", ";
		}
		if(std::isnan(num.y())){
			text += "Z: -nan";
		} else {
			text += "Z: " + std::to_string(num.z());
		}
		return text;
	}

	/* Description: Check for NAN.
	 * Argument: cv::Mat frame: Frame taken from the camera.
	 */
	void imagePublisher(cv::Mat frame)
	{
		sensor_msgs::ImagePtr msg = cv_bridge::CvImage(std_msgs::Header(), sensor_msgs::image_encodings::BGR8, frame).toImageMsg();
		image_pub_.publish(msg);
	}
};

int main(int argc, char* argv[]){
	ros::init(argc, argv, "image_publisher");

	cv::VideoCapture cap(0);
	cv::Mat frame;
	arMarkerRecognizer ar;
	const char* str = "true";

	if(argc > 1){
		if(strcmp(argv[1], str) == 0){
			ar.mode = true;
		}
	}
	if(!cap.isOpened()){
		ROS_INFO("Failed to open camera.");
		return -1;
	}

    if(cap.read(frame)){
    	ar.h_size = frame.size().width;
	    ar.v_size = frame.size().height;
    	if(ar.h_size == 0 && ar.v_size == 0){
	    	ar.h_size = 1280;
	    	ar.v_size = 720;
    	}
    }
	ar.cameraInternalParam(ar.mode);
	ros::Rate rate(100);

	while(ar.nh_.ok()){
		if(cap.read(frame)){
			ar.recognizeArMarker(frame);
			ar.imagePublisher(frame);
		}

		ros::spinOnce();
		rate.sleep();
	}

	cap.release();
	cv::destroyAllWindows();

	return 0;
}
